import unittest

from ruleengine.dsl.validator import validate_condition, validate_action


class ValidatorTest(unittest.TestCase):
    def test_condition_validation(self):
        validate_condition("123")
        validate_condition("'hello'")
        validate_condition("topic_is('blah')")
        validate_condition("and_(topic_is('blah'), msg.something == 123)")

        with self.assertRaises(Exception):
            validate_condition("")

        with self.assertRaises(Exception):
            validate_condition("missing_function(123)")

        with self.assertRaises(Exception):
            # Actions should not be part of condition
            validate_condition("create_moment('hello')")

    def test_action_validation(self):
        validate_action("create_moment('hello')")
        validate_action("create_moment('hello', description='', duration=100)")
        validate_action("create_moment(msg.title, description='', duration=100)")
        validate_action("upload(title='hello', description='', before=1)")

        with self.assertRaises(Exception):
            # Mispelled action name
            validate_action("create_momen('hello')")

        with self.assertRaises(Exception):
            # Wrong keyword arg
            validate_action("create_moment('hello', descrin='', durion=100)")



import re

from .condition import Condition, ThunkCondition

always = Condition.wrap(True)
identity = ThunkCondition(lambda item, scope: (item, scope))
msg = identity.msg
ts = identity.ts
topic = identity.topic
msgtype = identity.msgtype


def and_(*conditions):
    assert len(conditions) > 0, "and_ must have at least 1 condition"

    def new_thunk(item, scope):
        value = True
        for condition in conditions:
            value, scope = Condition.wrap(condition).evaluate_condition_at(item, scope)
            if not value:
                break
        return value, scope

    return ThunkCondition(new_thunk)


def or_(*conditions):
    assert len(conditions) > 0, "or_ must have at least 1 condition"

    def new_thunk(item, scope):
        value = False
        for condition in conditions:
            value, scope = Condition.wrap(condition).evaluate_condition_at(item, scope)
            if value:
                break
        return value, scope

    return ThunkCondition(new_thunk)


def not_(condition):
    return Condition.wrap(condition).map_condition_value(lambda x: not x)


def is_none(condition):
    def new_thunk(item, scope):
        value, scope = Condition.wrap(condition).evaluate_condition_at(item, scope)
        return value is None, scope

    return ThunkCondition(new_thunk)


def concat(*pieces):
    def new_thunk(item, scope):
        str_pieces = []
        for p in pieces:
            value, scope = Condition.wrap(p).evaluate_condition_at(item, scope)
            str_pieces.append(str(value))
        return "".join(str_pieces), scope

    return ThunkCondition(new_thunk)


def get_value(key):
    return Condition.wrap(key).map_condition_value(
        lambda k: ThunkCondition(lambda item, scope: (scope[k], scope))
    )


def set_value(key, value):
    return Condition.wrap(key).map_condition_value(
        lambda actual_key: Condition.wrap(value).map_condition_value(
            lambda actual_value: ThunkCondition(
                lambda item, scope: (True, {**scope, actual_key: actual_value})
            )
        )
    )


def topic_is(name):
    return topic == name


def type_is(name):
    return msgtype == name


def has(parent, child):
    return Condition.wrap(parent).map_condition_value(
        lambda p: Condition.wrap(child).map_condition_value(
            lambda c: ThunkCondition(
                lambda item, scope: (
                    c in p,
                    {**scope, "cos/contains": c if c in p else None},
                )
            )
        )
    )


def regex_search(value, pattern):
    return (
        Condition.wrap(value)
        .map_condition_value(lambda v: re.search(pattern, v))
        .map_condition_value(
            lambda match_result: ThunkCondition(
                lambda item, scope: (match_result, {**scope, "cos/regex": match_result})
            )
        )
    )


__all__ = [
    "always",
    "identity",
    "msg",
    "ts",
    "topic",
    "msgtype",
    "and_",
    "or_",
    "not_",
    "concat",
    "get_value",
    "set_value",
    "topic_is",
    "type_is",
    "has",
    "regex_search",
    "is_none",
]



from ruleengine.dsl.validation.validator import validate_action, validate_condition
from ruleengine.dsl.validation.validation_result import ValidationErrorType

ALLOWED_VERSIONS = ["1", "1.0.0", "v1"]


def validate_config(config, action_impls):
    """
    Validate a rule specification.

    It is assumed that the input config is a DiagnosisRule specified in

    https://github.com/coscene-io/cosceneapis/blob/main/coscene/dataplatform/v1alpha2/common/diagnosis_rule.proto

    and this function will return a DiagnosisRuleValidationResult specified in

    https://github.com/coscene-io/cosceneapis/blob/main/coscene/dataplatform/v1alpha2/common/diagnosis_rule_validation_result.proto
    """

    # TODO: Instead of putting together these objects by hand, we should connect
    # the protos generated in cosceneapis

    if config["version"] not in ALLOWED_VERSIONS:
        return {
            "success": False,
            "errors": [{"unexpectedVersion": {"allowedVersions": ALLOWED_VERSIONS}}],
        }

    errors = []
    rules = []
    for i, rule in enumerate(config["rules"]):
        rule_errors, conditions, actions = _validate_rule(rule, i, action_impls)
        errors += rule_errors
        rules.append((conditions, actions))

    success = not bool(errors)
    return {"success": success, "errors": errors}, rules if success else None


def _validate_rule(rule, rule_index, action_impls):
    errors = []
    if not rule["when"]:
        errors.append(
            {
                "location": {
                    "ruleIndex": rule_index,
                    "section": 1,
                },
                "emptySection": {},
            }
        )
    if not rule["actions"]:
        errors.append(
            {
                "location": {
                    "ruleIndex": rule_index,
                    "section": 2,
                },
                "emptySection": {},
            }
        )

    conditions = []
    for i, cond_str in enumerate(rule["when"]):
        res = validate_condition(cond_str)
        if not res.success:
            errors.append(
                {
                    "location": {
                        "ruleIndex": rule_index,
                        "section": 1,
                        "itemIndex": i,
                    },
                    **_convert_to_json_error(res),
                }
            )
        else:
            conditions.append(res.entity)

    actions = []
    for i, action_str in enumerate(rule["actions"]):
        res = validate_action(action_str, action_impls)
        if not res.success:
            errors.append(
                {
                    "location": {
                        "ruleIndex": rule_index,
                        "section": 2,
                        "itemIndex": i,
                    },
                    **_convert_to_json_error(res),
                }
            )
        else:
            actions.append(res.entity)
    return errors, conditions, actions


def _convert_to_json_error(result):
    match result.error_type:
        case ValidationErrorType.SYNTAX | ValidationErrorType.EMPTY:
            return {"syntax_error": {}}

        case ValidationErrorType.NOT_CONDITION:
            return {"notCondition": {"actualType": result.details["actual"]}}
        case ValidationErrorType.NOT_ACTION:
            return {"notAction": {"actualType": result.details["actual"]}}

        case ValidationErrorType.UNDEFINED:
            return {"nameUndefined": {"name": result.details["name"]}}

        case ValidationErrorType.UNKNOWN:
            return {"genericError": {"msg": result.details["message"]}}
        case _:
            raise Exception(f"Unknown error type: {result.error_type}")



# TODO: This validator was written in a rush. It has a number of issues, and is
# now deprecated. After switching over to the `ruleengine.dsl.validation`
# package, we should remove this file.

import inspect
from ruleengine.dsl import base_conditions, log_conditions, sequence_conditions


base_dsl_values = dict(
    inspect.getmembers(base_conditions)
    + inspect.getmembers(log_conditions)
    + inspect.getmembers(sequence_conditions)
)


def upload(
    before=10,
    title="",
    description="",
    labels=[],
    extra_files=[],
):
    pass


def create_moment(
    title,
    description="",
    timestamp=0,
    duration=1,
    create_task=False,
    assign_to=None,
):
    pass


actions_dsl_values = {
    "upload": upload,
    "create_moment": create_moment,
    **base_dsl_values,
}


def validate_condition(cond_str):
    # TODO: Very much not safe. Condition strings are user supplied, and we need
    # to sanitize the fuck out of it before doing eval.
    return eval(cond_str, base_dsl_values)


def validate_action(action_str):
    # TODO: Very much not safe. Condition strings are user supplied, and we need
    # to sanitize the fuck out of it before doing eval.
    return eval(action_str, actions_dsl_values)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Validate condition or action strings")
    parser.add_argument(
        "mode",
        help="What we're validating, action or condition",
        choices=["action", "condition"],
    )
    parser.add_argument("content", help="Content string to be validated")
    args = parser.parse_args()

    match args.mode:
        case "action":
            validate_action(args.content)
        case "condition":
            validate_condition(args.content)



import operator as op
from abc import ABC, abstractmethod


class Condition(ABC):
    """
    Defines a condition to be evaluated on a single message.

    It is assumed that the message has three fields, mirroring those of the ROS
    bag reader API:

    - topic
    - msg
    - msgtype
    - ts

    To make the DSL look nice at the end, this class ends up taking on a lot of
    the complexity. If you're familiar with monads, this class is a combination
    of Future and State and Optional. `map_condition_value` is both map and
    flatmap, and `wrap` is pure.

    If you're not familiar with monads, this is going to be very confusing. I'm
    sorry.

    """

    @abstractmethod
    def evaluate_condition_at(self, item, scope):
        pass

    @staticmethod
    def wrap(value):
        if isinstance(value, Condition):
            return value
        return ThunkCondition(lambda item, scope: (value, scope))

    def map_condition_value(self, mapper):
        def new_thunk(item, scope):
            value1, scope = self.evaluate_condition_at(item, scope)
            if value1 is None:
                return value1, scope

            mapped = mapper(value1)
            if not isinstance(mapped, Condition):
                return mapped, scope

            return mapped.evaluate_condition_at(item, scope)

        return ThunkCondition(new_thunk)

    def __eq__(self, other):
        return self.__wrap_binary_op(other, op.eq)

    def __ne__(self, other):
        return self.__wrap_binary_op(other, op.ne)

    def __gt__(self, other):
        return self.__wrap_binary_op(other, op.gt, float)

    def __ge__(self, other):
        return self.__wrap_binary_op(other, op.ge, float)

    def __lt__(self, other):
        return self.__wrap_binary_op(other, op.lt, float)

    def __le__(self, other):
        return self.__wrap_binary_op(other, op.le, float)

    def __call__(self, *args, **kwargs):
        return self.map_condition_value(lambda f: f(*args, **kwargs))

    def __getattr__(self, name):
        return self.map_condition_value(lambda x: getattr(x, name, None))

    def __wrap_binary_op(self, other, op, coerce=lambda x: x):
        return self.map_condition_value(
            lambda x: Condition.wrap(other).map_condition_value(
                lambda y: op(coerce(x), coerce(y))
            )
        )

    def __bool__(self):
        pass
        raise NotImplementedError(
            """
            It is intentional that Condition objects should not be used as boolean values.

            If you're trying to use boolean operators with conditions, please use function equivalents instead:
              a and b -> and_(a, b)
              a or b -> or_(a, b)
              not a -> not_(a)

            If you're trying to use the `in` operator, please also use function equivalent instead:
              a in b -> has(b, a)
            """
        )


class ThunkCondition(Condition):
    def __init__(self, thunk):
        super().__init__()
        self.__thunk = thunk

    def evaluate_condition_at(self, item, scope):
        return self.__thunk(item, scope)


__all__ = [
    "Condition",
    "ThunkCondition",
]



from .dsl.action import Action
from .dsl.condition import Condition

# TODO: Rule and Engine classes are deprecated, remove the code after cleaning
# up call sites. New code should use `run` below.


class Rule:
    def __init__(self, condition, action):
        assert isinstance(condition, Condition)
        assert isinstance(action, Action)

        self.__cond = condition
        self.__action = action

    def eval_condition(self, item):
        return self.__cond.evaluate_condition_at(item, {})

    def run_action(self, item, scope):
        return self.__action.run(item, scope)


class Engine:
    def __init__(self, rules, data):
        self.__rules = rules
        self.__data = data

    def run(self):
        for item in self.__data:
            for rule in self.__rules:
                result, scope = rule.eval_condition(item)
                if result:
                    rule.run_action(item, scope)


class Engine2:
    def __init__(self, rules):
        self.__rules = rules

    def consume_next(self, item):
        for conditions, actions in self.__rules:
            for cond in conditions:
                res, scope = cond.evaluate_condition_at(item, {})
                if res:
                    for action in actions:
                        action.run(item, scope)
                    break
